#!/usr/bin/python
# QLBR Charm
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ConfigParser import (
    NoOptionError,
    SafeConfigParser,
)
import os
import sys

from base64 import b64decode
from subprocess import check_call

from charmhelpers.core import (
    hookenv,
    host,
)
from charmhelpers import fetch
from charmhelpers.fetch.bzrurl import BzrUrlFetchHandler


hooks = hookenv.Hooks()
log = hookenv.log
config = hookenv.config()

SERVICE = 'gunicorn'
MANAGE = os.path.join(config['qlbr_dir'], 'manage.py')
MANAGE_COMMAND = ['sudo', '-u', config['wsgi_user'], MANAGE]
DB_FILE = os.path.join(config['qlbr_dir'], 'qlbr', 'database_settings.py')
CACHE_FILE = os.path.join(config['qlbr_dir'], 'qlbr', 'cache_settings.py')

CRON_TEMPLATE = (
    '0 * * * * {user} '
    '{manage} get_task_data reload --credentials_file={credentials}\n'
    '20,30,40,50 * * * * {user} '
    '{manage} get_task_data --credentials_file={credentials}\n'
    '35 0 * * * {user} '
    '{manage} update_people --credentials_file={credentials}\n'
)


def setup_cron():
    credentials = config['credentials_file']
    cron_file_path = config['cron_file']
    with open(cron_file_path, 'w') as cron_file:
        cron_file.write(
            CRON_TEMPLATE.format(
                manage=MANAGE,
                credentials=credentials,
                user=config['wsgi_user'],
            )
        )
    os.chmod(cron_file_path, 0755)


def write_credentials_file():
    file_name = config['credentials_file']
    credentials = b64decode(config['credentials'])
    with open(file_name, 'w') as credentials_file:
        credentials_file.write(credentials)
    parser = SafeConfigParser()
    parser.read(file_name)
    keys = (
        'consumer_key',
        'consumer_secret',
        'access_token',
        'access_secret',
    )
    for key in keys:
        try:
            parser.get('1', key)
        except NoOptionError:
            raise KeyError(
                '{} missing from credentials; cannot connect to launchpad'
                .format(key))


@hooks.hook('config-changed')
def config_changed():
    for key in config:
        if config.changed(key):
            log("config['{}'] changed from {} to {}".format(
                key, config.previous(key), config[key]))

    config.save()
    start()


@hooks.hook('start')
def start():
    host.service_restart(SERVICE) or host.service_start(SERVICE)


@hooks.hook('stop')
def stop():
    host.service_stop(SERVICE)


@hooks.hook('install')
def install():
    wu = config(['wsgi_user'])
    wg = config(['wsgi_group'])
    qu = config(['qlbr_user'])
    qg = config(['qlbr_group'])
    host.add_user(wu)
    host.add_user_to_group(wu, wg)
    host.add_user(qu)
    host.add_user_to_group(qu, qg)

    for ppa in config['ppas'].split():
        fetch.add_source(ppa)
    fetch.apt_update()
    fetch.apt_install(config['distro_packages'].split())
    handler = BzrUrlFetchHandler()
    qlbr_dir = config['qlbr_dir']
    handler.branch(config['qlbr_branch'], qlbr_dir)
    write_credentials_file()
    host.chownr(qlbr_dir, qu, qg)


@hooks.hook('pgsql-relation-joined', 'pgsql-relation-changed')
def pgsql_relation_joined_changed():
    # Cribbed this check from the python-django charm
    database = hookenv.relation_get('database')
    if not database:
        return True

    user = hookenv.relation_get('user')
    password = hookenv.relation_get('password')
    host = hookenv.relation_get('host')
    port = 5432

    settings = '''DATABASES = {{
    'default': {{
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': '{database}',
        'USER': '{user}',
        'PASSWORD': '{password}',
        'HOST': '{host}',
        'PORT': '{port}',
        'OPTIONS': {{'autocommit': True}},
    }}
}}'''.format(
        database=database,
        user=user,
        password=password,
        host=host,
        port=port,
    )

    with open(DB_FILE, 'w') as settings_file:
        settings_file.write(settings)

    qlbr_dir = config['qlbr_dir']
    searches = os.path.join(qlbr_dir, 'qlbr', 'searches.json')
    check_call(MANAGE_COMMAND + ['syncdb', '--noinput'])
    check_call(MANAGE_COMMAND + ['flush', '--noinput'])
    check_call(MANAGE_COMMAND + ['load_searches', searches])

    setup_cron()


@hooks.hook('pgsql-relation-broken')
def pgsql_relation_broken():
    os.unlink(DB_FILE)
    os.unlink(config['cron_file'])


@hooks.hook('wsgi-file-relation-joined', 'wsgi-file-relation-changed')
def wsgi_file_relation_joined_changed():
    hookenv.relation_set(
        working_dir=config['qlbr_dir'],
        port=config['port'],
        wsgi_user=config['wsgi_user'],
        wsgi_group=config['wsgi_group'],
    )
    hookenv.open_port(config['port'])
    start()


@hooks.hook('wsgi-file-relation-broken')
def wsgi_file_relation_broken():
    hookenv.close_port(config['port'])
    stop()


@hooks.hook('memcached-relation-joined', 'memcached-relation-changed')
def memcached_relation_joined_changed():
    # Cribbed this check from the python-django charm
    host = hookenv.relation_get('host')
    if not host:
        return True

    port = hookenv.relation_get('port')

    settings = '''CACHES = {{
    'default': {{
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': '{host}:{port}',
    }}
}}'''.format(host=host, port=port)

    with open(CACHE_FILE, 'w') as settings_file:
        settings_file.write(settings)

    start()


@hooks.hook('memcached-relation-broken')
def memcached_relation_broken():
    os.unlink(CACHE_FILE)
    start()


if __name__ == "__main__":
    # execute a hook based on the name the program is called by
    hooks.execute(sys.argv)
